@@ -20,10 +20,12 @@ from oslo_serialization import jsonutils
 from oslo_utils import strutils
 import six
 from six.moves.urllib import request
+import time
 import yaml
 
 from heatclient.common import deployment_utils
 from heatclient.common import event_utils
+from heatclient.common import http
 from heatclient.common import template_format
 from heatclient.common import template_utils
 from heatclient.common import utils
@@ -41,7 +43,11 @@ def _authenticated_fetcher(hc):
     """A wrapper around the heat client object to fetch a template.
     """
     def _do(*args, **kwargs):
-        return hc.http_client.raw_request(*args, **kwargs).content
+        if isinstance(hc.http_client, http.SessionClient):
+            method, url = args
+            return hc.http_client.request(url, method, **kwargs).content
+        else:
+            return hc.http_client.raw_request(*args, **kwargs).content
 
     return _do
 
@@ -79,13 +85,19 @@ def _authenticated_fetcher(hc):
            'This can be specified multiple times, or once with parameters '
            'separated by a semicolon.'),
            action='append')
-@utils.arg('-Pf', '--parameter-file', metavar='<KEY=VALUE>',
+@utils.arg('-Pf', '--parameter-file', metavar='<KEY=FILE>',
            help=_('Parameter values from file used to create the stack. '
            'This can be specified multiple times. Parameter value '
            'would be the content of the file'),
            action='append')
+@utils.arg('--poll', metavar='SECONDS', type=int, nargs='?', const=5,
+           help=_('Poll and report events until stack completes. '
+                  'Optional poll interval in seconds can be provided as '
+                  'argument, default 5.'))
 @utils.arg('name', metavar='<STACK_NAME>',
            help=_('Name of the stack to create.'))
+@utils.arg('--tags', metavar='<TAG1,TAG2>',
+           help=_('A list of tags to associate with the stack.'))
 def do_stack_create(hc, args):
     '''Create the stack.'''
     tpl_files, template = template_utils.get_template_contents(
@@ -119,12 +131,16 @@ def do_stack_create(hc, args):
         'environment': env
     }
 
+    if args.tags:
+        fields['tags'] = args.tags
     timeout = args.timeout or args.create_timeout
     if timeout:
         fields['timeout_mins'] = timeout
 
     hc.stacks.create(**fields)
     do_stack_list(hc)
+    if args.poll is not None:
+        _poll_for_events(hc, args.name, 'CREATE', poll_period=args.poll)
 
 
 def hooks_to_env(env, arg_hooks, hook):
@@ -234,13 +250,15 @@ def do_stack_adopt(hc, args):
            'This can be specified multiple times, or once with parameters '
            'separated by semicolon.'),
            action='append')
-@utils.arg('-Pf', '--parameter-file', metavar='<KEY=VALUE>',
+@utils.arg('-Pf', '--parameter-file', metavar='<KEY=FILE>',
            help=_('Parameter values from file used to create the stack. '
            'This can be specified multiple times. Parameter value '
            'would be the content of the file'),
            action='append')
 @utils.arg('name', metavar='<STACK_NAME>',
            help=_('Name of the stack to preview.'))
+@utils.arg('--tags', metavar='<TAG1,TAG2>',
+           help=_('A list of tags to associate with the stack.'))
 def do_stack_preview(hc, args):
     '''Preview the stack.'''
     tpl_files, template = template_utils.get_template_contents(
@@ -264,6 +282,9 @@ def do_stack_preview(hc, args):
         'environment': env
     }
 
+    if args.tags:
+        fields['tags'] = args.tags
+
     stack = hc.stacks.preview(**fields)
     formatters = {
         'description': utils.text_wrap_formatter,
@@ -371,20 +392,7 @@ def do_action_check(hc, args):
 def do_stack_show(hc, args):
     '''Describe the stack.'''
     fields = {'stack_id': args.id}
-    try:
-        stack = hc.stacks.get(**fields)
-    except exc.HTTPNotFound:
-        raise exc.CommandError(_('Stack not found: %s') % args.id)
-    else:
-        formatters = {
-            'description': utils.text_wrap_formatter,
-            'template_description': utils.text_wrap_formatter,
-            'stack_status_reason': utils.text_wrap_formatter,
-            'parameters': utils.json_formatter,
-            'outputs': utils.json_formatter,
-            'links': utils.link_formatter
-        }
-        utils.print_dict(stack.to_dict(), formatters=formatters)
+    _do_stack_show(hc, fields)
 
 
 @utils.arg('-f', '--template-file', metavar='<FILE>',
@@ -425,7 +433,7 @@ def do_stack_show(hc, args):
            'This can be specified multiple times, or once with parameters '
            'separated by a semicolon.'),
            action='append')
-@utils.arg('-Pf', '--parameter-file', metavar='<KEY=VALUE>',
+@utils.arg('-Pf', '--parameter-file', metavar='<KEY=FILE>',
            help=_('Parameter values from file used to create the stack. '
            'This can be specified multiple times. Parameter value '
            'would be the content of the file'),
@@ -444,6 +452,8 @@ def do_stack_show(hc, args):
            action='append')
 @utils.arg('id', metavar='<NAME or ID>',
            help=_('Name or ID of stack to update.'))
+@utils.arg('--tags', metavar='<TAG1,TAG2>',
+           help=_('An updated list of tags to associate with the stack.'))
 def do_stack_update(hc, args):
     '''Update the stack.'''
 
@@ -471,6 +481,8 @@ def do_stack_update(hc, args):
         'environment': env
     }
 
+    if args.tags:
+        fields['tags'] = args.tags
     if args.timeout:
         fields['timeout_mins'] = args.timeout
     if args.clear_parameter:
@@ -517,6 +529,18 @@ def do_stack_cancel_update(hc, args):
            'This can be specified multiple times, or once with parameters '
            'separated by a semicolon.'),
            action='append')
+@utils.arg('-t', '--tags', metavar='<TAG1,TAG2...>',
+           help=_('Show stacks containing these tags, combine multiple tags '
+                  'using the boolean AND expression'))
+@utils.arg('--tags-any', metavar='<TAG1,TAG2...>',
+           help=_('Show stacks containing these tags, combine multiple tags '
+                  'using the boolean OR expression'))
+@utils.arg('--not-tags', metavar='<TAG1,TAG2...>',
+           help=_('Show stacks not containing these tags, combine multiple '
+                  'tags using the boolean AND expression'))
+@utils.arg('--not-tags-any', metavar='<TAG1,TAG2...>',
+           help=_('Show stacks not containing these tags, combine multiple '
+                  'tags using the boolean OR expression'))
 @utils.arg('-l', '--limit', metavar='<LIMIT>',
            help=_('Limit the number of stacks returned.'))
 @utils.arg('-m', '--marker', metavar='<ID>',
@@ -535,6 +559,10 @@ def do_stack_list(hc, args=None):
         kwargs = {'limit': args.limit,
                   'marker': args.marker,
                   'filters': utils.format_parameters(args.filters),
+                  'tags': args.tags,
+                  'tags_any': args.tags_any,
+                  'not_tags': args.not_tags,
+                  'not_tags_any': args.not_tags_any,
                   'global_tenant': args.global_tenant,
                   'show_deleted': args.show_deleted,
                   'show_hidden': args.show_hidden}
@@ -707,11 +735,15 @@ def do_template_validate(hc, args):
            help=_('Name or ID of stack to show the resources for.'))
 @utils.arg('-n', '--nested-depth', metavar='<DEPTH>',
            help=_('Depth of nested stacks from which to display resources.'))
+@utils.arg('--with-detail', default=False, action="store_true",
+           help=_('Enable detail information presented for each resource '
+                  'in resources list.'))
 def do_resource_list(hc, args):
     '''Show list of resources belonging to a stack.'''
     fields = {
         'stack_id': args.id,
         'nested_depth': args.nested_depth,
+        'with_detail': args.with_detail,
     }
     try:
         resources = hc.resources.list(**fields)
@@ -725,8 +757,8 @@ def do_resource_list(hc, args):
         else:
             fields.insert(0, 'resource_name')
 
-        if args.nested_depth:
-            fields.append('parent_resource')
+        if args.nested_depth or args.with_detail:
+            fields.append('stack_name')
 
         utils.print_list(resources, fields, sortby_index=4)
 
@@ -735,10 +767,16 @@ def do_resource_list(hc, args):
            help=_('Name or ID of stack to show the resource for.'))
 @utils.arg('resource', metavar='<RESOURCE>',
            help=_('Name of the resource to show the details for.'))
+@utils.arg('-a', '--with-attr', metavar='<ATTRIBUTE>',
+           help=_('Attribute to show, it can be specified '
+                  'multiple times.'),
+           action='append')
 def do_resource_show(hc, args):
     '''Describe the resource.'''
     fields = {'stack_id': args.id,
               'resource_name': args.resource}
+    if args.with_attr:
+        fields['with_attr'] = list(args.with_attr)
     try:
         resource = hc.resources.get(**fields)
     except exc.HTTPNotFound:
@@ -897,6 +935,9 @@ def do_hook_clear(hc, args):
 @utils.arg('-n', '--nested-depth', metavar='<DEPTH>',
            help=_('Depth of nested stacks from which to display events. '
                   'Note this cannot be specified with --resource.'))
+@utils.arg('-F', '--format', metavar='<FORMAT>',
+           help=_('The output value format, one of: log, table'),
+           default='table')
 def do_event_list(hc, args):
     '''List events for a stack.'''
     display_fields = ['id', 'resource_status_reason',
@@ -937,7 +978,10 @@ def do_event_list(hc, args):
         else:
             display_fields.insert(0, 'logical_resource_id')
 
-    utils.print_list(events, display_fields, sortby_index=None)
+    if args.format == 'log':
+        print(utils.event_log_formatter(events))
+    else:
+        utils.print_list(events, display_fields, sortby_index=None)
 
 
 def _get_hook_type_via_status(hc, stack_id):
@@ -1115,7 +1159,7 @@ def do_config_show(hc, args):
 @utils.arg('id', metavar='<ID>', nargs='+',
            help=_('IDs of the configurations to delete.'))
 def do_config_delete(hc, args):
-    '''Delete software configurations.'''
+    '''Delete a software configuration.'''
     failure_count = 0
 
     for config_id in args.id:
@@ -1139,7 +1183,7 @@ def do_config_delete(hc, args):
                   'CREATE, UPDATE, DELETE, SUSPEND, RESUME'))
 @utils.arg('-c', '--config', metavar='<CONFIG>',
            help=_('ID of the configuration to deploy.'))
-@utils.arg('-s', '--server', metavar='<SERVER>',
+@utils.arg('-s', '--server', metavar='<SERVER>', required=True,
            help=_('ID of the server being deployed to.'))
 @utils.arg('-t', '--signal-transport',
            default='TEMP_URL_SIGNAL',
@@ -1162,10 +1206,14 @@ def do_config_delete(hc, args):
                   'deployment. This is used to apply a sort order to the '
                   'list of configurations currently deployed to the server.'))
 def do_deployment_create(hc, args):
-    try:
-        config = hc.software_configs.get(config_id=args.config)
-    except exc.HTTPNotFound:
-        raise exc.CommandError(_('Configuration not found: %s') % args.id)
+    '''Create a software deployment.'''
+    config = {}
+    if args.config:
+        try:
+            config = hc.software_configs.get(config_id=args.config)
+        except exc.HTTPNotFound:
+            raise exc.CommandError(
+                _('Configuration not found: %s') % args.config)
 
     derrived_params = deployment_utils.build_derived_config_params(
         action=args.action,
@@ -1188,6 +1236,17 @@ def do_deployment_create(hc, args):
     print(jsonutils.dumps(sd.to_dict(), indent=2))
 
 
+@utils.arg('-s', '--server', metavar='<SERVER>',
+           help=_('ID of the server to fetch deployments for.'))
+def do_deployment_list(hc, args):
+    '''List software deployments.'''
+    kwargs = {'server_id': args.server} if args.server else {}
+    deployments = hc.software_deployments.list(**kwargs)
+    fields = ['id', 'config_id', 'server_id', 'action', 'status',
+              'creation_time', 'status_reason']
+    utils.print_list(deployments, fields, sortby_index=5)
+
+
 @utils.arg('id', metavar='<ID>',
            help=_('ID of the deployment.'))
 def do_deployment_show(hc, args):
@@ -1211,7 +1270,7 @@ def do_deployment_metadata_show(hc, args):
 @utils.arg('id', metavar='<ID>', nargs='+',
            help=_('IDs of the deployments to delete.'))
 def do_deployment_delete(hc, args):
-    '''Delete software deployments.'''
+    '''Delete a software deployment.'''
     failure_count = 0
 
     for deploy_id in args.id:
@@ -1235,7 +1294,7 @@ def do_deployment_delete(hc, args):
            help=_('The output value format, one of: raw, json'),
            default='raw')
 def do_deployment_output_show(hc, args):
-    '''Show a specific stack output.'''
+    '''Show a specific deployment output.'''
     if (not args.all and args.output is None or
             args.all and args.output is not None):
         raise exc.CommandError(
@@ -1342,15 +1401,12 @@ def do_snapshot_list(hc, args):
     except exc.HTTPNotFound:
         raise exc.CommandError(_('Stack not found: %s') % args.id)
     else:
-        fields = ['id', 'name', 'status', 'status_reason', 'data',
-                  'creation_time']
+        fields = ['id', 'name', 'status', 'status_reason', 'creation_time']
         formatters = {
             'id': lambda x: x['id'],
             'name': lambda x: x['name'],
             'status': lambda x: x['status'],
             'status_reason': lambda x: x['status_reason'],
-            'data': lambda x: jsonutils.dumps(x['data'], indent=2,
-                                              ensure_ascii=False),
             'creation_time': lambda x: x['creation_time'],
         }
         utils.print_list(snapshots["snapshots"], fields, formatters=formatters)
@@ -1362,3 +1418,75 @@ def do_service_list(hc, args=None):
               'topic', 'updated_at', 'status']
     services = hc.services.list()
     utils.print_list(services, fields, sortby_index=1)
+
+
+def do_template_version_list(hc, args):
+    '''List the available template versions.'''
+    versions = hc.template_versions.list()
+    fields = ['version', 'type']
+    utils.print_list(versions, fields, sortby_index=1)
+
+
+@utils.arg('template_version', metavar='<TEMPLATE_VERSION>',
+           help=_('Template version to get the functions for.'))
+def do_template_function_list(hc, args):
+    '''List the available functions.'''
+    try:
+        functions = hc.template_versions.get(args.template_version)
+    except exc.HTTPNotFound:
+        raise exc.CommandError(
+            _('Template version not found: %s') % args.template_version)
+    else:
+        utils.print_list(functions, ['functions', 'description'])
+
+
+def _do_stack_show(hc, fields):
+    try:
+        stack = hc.stacks.get(**fields)
+    except exc.HTTPNotFound:
+        raise exc.CommandError(_('Stack not found: %s') %
+                               fields.get('stack_id'))
+    else:
+        formatters = {
+            'description': utils.text_wrap_formatter,
+            'template_description': utils.text_wrap_formatter,
+            'stack_status_reason': utils.text_wrap_formatter,
+            'parameters': utils.json_formatter,
+            'outputs': utils.json_formatter,
+            'links': utils.link_formatter
+        }
+        utils.print_dict(stack.to_dict(), formatters=formatters)
+
+
+def _poll_for_events(hc, stack_name, action, poll_period):
+    """When an action is performed on a stack, continuously poll for its
+    events and display to user as logs.
+    """
+    fields = {'stack_id': stack_name}
+    _do_stack_show(hc, fields)
+    marker = None
+    while True:
+        events = event_utils.get_events(hc, stack_id=stack_name,
+                                        event_args={'sort_dir': 'asc',
+                                                    'marker': marker})
+
+        if len(events) >= 1:
+            # set marker to last event that was received.
+            marker = getattr(events[-1], 'id', None)
+            events_log = utils.event_log_formatter(events)
+            print(events_log)
+            for event in events:
+                # check if stack event was also received
+                if getattr(event, 'resource_name', '') == stack_name:
+                    stack_status = getattr(event, 'resource_status', '')
+                    msg = _("\n Stack %(name)s %(status)s \n") % dict(
+                        name=stack_name, status=stack_status)
+                    if stack_status == '%s_COMPLETE' % action:
+                        _do_stack_show(hc, fields)
+                        print(msg)
+                        return
+                    elif stack_status == '%s_FAILED' % action:
+                        _do_stack_show(hc, fields)
+                        raise exc.StackFailure(msg)
+
+        time.sleep(poll_period)
